import java.awt.*;
import java.util.*;

/**An element is the superclass of all the graphical objects in a
 * presentation. Each element has a number of positions at different times
 * in the timeline (to be tweened between for other times).
 */
abstract class Element
{
	Vector pos = new Vector();
	Vector acts = new Vector();
	String name;            // unique name for this Element
	Color fcol, bcol;       // foreground and background colors
	TimeLine parent;        // timeline element is in
	int idx;                // index in element list in TimeLine
	int x,y;                // computed position for rendering
	boolean visible;        // computed visibility
	ElementEditor locked;   // who this element is being edited by

	/**Add a new tweening point at the correct position
	 */
	void addPosition(ElementPos p)
	{
	for(int i=0; i<pos.size(); i++)
		if (p.time < ((ElementPos)pos.elementAt(i)).time) {
			for(int j=i; j<pos.size(); j++)
				((ElementPos)pos.elementAt(j)).idx++;
			pos.insertElementAt(p, i);
			p.idx = i;
			return;
			}
	p.idx = pos.size();
	pos.addElement(p);
	}

	ElementPos getPosition(int i)
	{
	return (ElementPos)pos.elementAt(i);
	}

	void deletePosition(int i)
	{
	for(int j=i; j<pos.size(); j++)
		getPosition(j).idx--;
	pos.removeElementAt(i);
	}

	/**After an element object has been instantiated, this method will
	 * be called if the object is defined by a rectangle or line
	 * @param d		Size of the rectangle
	 * @param f		Foreground color
	 * @param b		Background color
	 */
	void create(Dimension d, Color f, Color b) { }

	/**After an element object has been instantiated, this method will
	 * be called if the object is defined by a list of points.
	 * @param p		Points, offset from position
	 * @param f		Foreground color
	 * @param b		Background color
	 */
	void create(Vector p, Color f, Color b) { }

	/**Returns 0 for elements bounded by a rectange, 1 for a line
	 * and 2 for a curve
	 */
	int getStyle()
	{
	return 0;
	}

	/**This function should return the current bounding box of this
	 * element (for elements where getStyle() returns 0)
	 */
	Rectangle getBounds() { return null; }

	/**This function should return the endpoints of the defining line
	 * (for elements where getStyle() returns 1)
	 */
	Point[] getEndPoints() { return null; }

	/**This method must be overridden to resize this element
	 */
	abstract void setSize(int nw, int nh);

	/**Render this element. Typically elements drawn in edit mode will
	 * be partly visible even if they should be hidden.
	 * @param g		Graphics to draw into
	 * @param e		Are we in edit mode?
	 */
	abstract void paint(Graphics g, boolean e);

	/**Copies this element into the one given. This is guaranteed
	 * to be an element of the same type.
	 * Subclasses should override this, but remember to call the super
	 * method.
	 */
	void copy(Element t)
	{
	t.name = name;
	t.pos = new Vector();
	t.fcol = fcol; t.bcol = bcol;
	for(int i=0; i<pos.size(); i++) {
		ElementPos ep = (ElementPos)pos.elementAt(i);
		t.pos.addElement(
		  new ElementPos(ep.time, ep.x, ep.y, ep.visible));
		}
	for(int i=0; i<acts.size(); i++) {
		Action a = (Action)acts.elementAt(i);
		t.acts.addElement(Action.loadAction(a.save()));
		}
	}

	/**Returns a new copy of this element, by calling the other copy
	 * method. Don't override this.
	 */
	Element copy()
	{
	Element t = (Element)Util.createObject(this);
	copy(t);
	return t;
	}

	/**Subclasses must override this to save themselves. They should call
	 * super.save() and add extra info into the returned HData object
	 */
	HData save()
	{
	HData hd = new HData();
	hd.addString("class", getClass().getName());
	hd.addString("name", name);
	if (fcol != null) hd.addColor("fcol", fcol);
	if (bcol != null) hd.addColor("bcol", bcol);
	Vector pv = new Vector();
	for(int i=0; i<pos.size(); i++)
		pv.addElement(((ElementPos)pos.elementAt(i)).save());
	hd.addVector("pos", pv);
	Vector av = new Vector();
	for(int i=0; i<acts.size(); i++)
		av.addElement(((Action)acts.elementAt(i)).save());
	hd.addVector("acts", av);
	return hd;
	}

	/**Subclasses must override this function to load their state
	 * from the given data object. super.load() must be called first
	 * to handle the superclass
	 */
	void load(HData hd)
	{
	name = hd.getString("name");
	Vector pv = hd.getVector("pos");
	if (hd.contains("fcol")) fcol = hd.getColor("fcol");
	if (hd.contains("bcol")) bcol = hd.getColor("bcol");
	for(int i=0; i<pv.size(); i++) {
		ElementPos ep = new ElementPos((HData)pv.elementAt(i));
		ep.idx = i;
		pos.addElement(ep);
		}
	Vector av = hd.getVector("acts");
	for(int i=0; i<av.size(); i++) {
		Action ac = Action.loadAction((HData)av.elementAt(i));
		ac.owner = this;
		acts.addElement(ac);
		}
	}

	boolean mouseDown(Event e, int x, int y)
	{
	return false;
	}

	boolean mouseUp(Event e, int x, int y)
	{
	return false;
	}

	boolean mouseDrag(Event e, int x, int y)
	{
	return false;
	}

	/**Sets the time in the timeline at which the element should be
	 * rendered.
	 * @param t		Time to render at
	 */
	void setTime(int t)
	{
	//Point p = tweenPoint(t);
	ElementPos pn[] = tweenPos(t), prv = pn[0], nxt = pn[1];
	visible = prv.visible;
	if (prv == nxt) {
		x = prv.x;
		y = prv.y;
		}
	else {
		x = prv.x+(nxt.x-prv.x)*(t-prv.time)/(nxt.time-prv.time);
		y = prv.y+(nxt.y-prv.y)*(t-prv.time)/(nxt.time-prv.time);
		}
	}

	/**Returns the element position at some time..
	 */
	protected Point tweenPoint(int t)
	{
	ElementPos pn[] = tweenPos(t), prv = pn[0], nxt = pn[1];
	if (prv == nxt)
		return new Point(prv.x, nxt.y);	
	return new Point(
	  prv.x + (nxt.x - prv.x)*(t - prv.time)/(nxt.time - prv.time),
	  prv.y + (nxt.y - prv.y)*(t - prv.time)/(nxt.time - prv.time));	
	}

	/**Returns the position objects before and afer the given time
	 * @param t		The time to find a position for
	 */
	ElementPos[] tweenPos(int t)
	{
	ElementPos rv[] = new ElementPos[2];
	for(int i=pos.size()-1; i>=0; i--) {
		ElementPos ep = (ElementPos)pos.elementAt(i);
		if (t >= ep.time) {
			rv[0] = ep;
			rv[1] = i==pos.size()-1 ? rv[0] :
			         (ElementPos)pos.elementAt(i+1);
			return rv;
			}
		}
	// time is before the 1st position..
	rv[0] = rv[1] = (ElementPos)pos.elementAt(0);
	return rv;
	}
}


class ElementPos
{
	int time;
	int x, y;
	boolean visible;
	int idx;

	ElementPos(int time, int x, int y, boolean visible)
	{
	this.time = time; this.x = x; this.y = y; this.visible = visible;
	}

	ElementPos(HData h)
	{
	this(h.getInteger("time"), h.getInteger("x"),
	     h.getInteger("y"), h.getBoolean("visible"));
	}

	HData save()
	{
	HData h = new HData();
	h.addInteger("time", time);
	h.addInteger("x", x); h.addInteger("y", y);
	h.addBoolean("visible", visible);
	return h;
	}
}
