import java.awt.*;
import java.util.*;

/**An element editor is a pop-up window for editing the attributes of
 * some element.
 */
abstract public class ElementEditor extends FixedFrame implements
	CbButtonCallback, ActionsWindowCallback
{
	/**Subclasses should put all their UI into this panel */
	Panel pan;
	ElementPosEditor posed;
	CbButton ok, apply, cancel, actions;
	TextField name;
	Element elm;
	EditArea edit_area;
	ActionsWindow actionswin;

	/**Create the basic element editor UI. Subclasses should override
	 * this method (but still call the superclass function)
	 */
	protected void makeUI()
	{
	setLayout(new BorderLayout());
	Panel bot = new GrayPanel();
	bot.setLayout(new FlowLayout(FlowLayout.RIGHT));
	bot.add(ok = new CbButton("Ok", this));
	bot.add(apply = new CbButton("Apply", this));
	bot.add(cancel = new CbButton("Cancel", this));
	add("South", bot);
	Panel mid = new GrayPanel();
	mid.setLayout(new BorderLayout());
	mid.add("East", posed = new ElementPosEditor(this));
	mid.add("West", pan = new BorderPanel(1));
	add("Center", mid);
	Panel top = new GrayPanel();
	top.setLayout(new BorderLayout());
	top.add("West", new Label("Name:"));
	top.add("Center", name = new TextField(elm.name));
	top.add("East", actions = new CbButton("Actions..", this));
	add("North", top);
	}

	/**Subclasses should override this method to copy element
	 * properties from input fields into the element being edited.
	 * Subclasses should check their inputs, call and check super.apply()
	 * and then apply their inputs.
	 * If some input is invalid, then the method should display an
	 * error window and return false.
	 */
	protected boolean apply()
	{
	Vector np = posed.getPositions();
	if (np == null) return false;
	String nm = name.getText();
	Element same = edit_area.pres.findElement(nm);
	if (same != null && same != elm) {
		new ErrorWindow("An element called "+nm+" already exists");
		return false;
		}
	elm.name = nm;
	elm.pos.removeAllElements();
	for(int i=0; i<np.size(); i++)
		elm.addPosition((ElementPos)np.elementAt(i));
	elm.setTime(edit_area.pres.ctl.time);
	return true;
	}

	/**Edit some element. Subclasses shouldn't override this
	 * @param el	The element to edit
	 * @param ea	Parent edit area
	 */
	void edit(Element el, EditArea ea)
	{
	elm = el; edit_area = ea;
	makeUI();
	Util.recursiveBackground(this, Color.lightGray);
	pack();
	show();
	setTitle("Edit "+el.getClass().getName());
	edit_area.lock(this, elm);
	}

	protected int getNum(TextField t)
	{
	try { return Integer.parseInt(t.getText()); }
	catch(Exception e) { return -Integer.MAX_VALUE; }
	}

	public void click(CbButton b)
	{
	if (b == ok || b == apply) {
		if (!apply())
			return;
		edit_area.repaint();
		}
	if (b == ok || b == cancel)
		dispose();
	if (b == actions && actionswin == null)
		actionswin = new ActionsWindow(elm.parent.parent, this, elm);
	}

	public void dispose()
	{
	super.dispose();
	edit_area.unlock(this, elm);
	}

	public void actionsWindowClosed(Object o)
	{
	actionswin = null;
	}

	public Vector listActions(Object o)
	{
	return elm.acts;
	}

	public void setActions(Object o, Vector a)
	{
	}

	public int[] getEvents(Object o)
	{
	int elist[] = { 1, 2, 3 };
	return elist;
	}
}

/**A component for editing the positions of an element.
 */
class ElementPosEditor extends BorderPanel implements CbScrollbarCallback,
						      CbButtonCallback
{
	Panel scr, midc;
	CbScrollbar sb;
	ElementEditor parent;

	ElementPosEditor(ElementEditor p)
	{
	super(1);
	parent = p;
	Element e = parent.elm;
	setLayout(new BorderLayout());
	add("East", sb = new CbScrollbar(CbScrollbar.VERTICAL, this));
	Panel mid = new GrayPanel();
	mid.setLayout(new BorderLayout());
	Panel midt = new GrayPanel();
	midt.setLayout(new GridLayout(1, 5));
	midt.add(new Label("Time"));
	midt.add(new Label("X"));
	midt.add(new Label("Y"));
	midt.add(new Label("Visible"));
	midt.add(new Label(""));
	mid.add("North", midt);
	midc = new GrayPanel();
	midc.setLayout(null);
	midc.add(scr = new GrayPanel());
	mid.add("Center", midc);
	add("Center", mid);
	scr.setLayout(new GridLayout(0, 5));

	// create initial position fields
	for(int i=0,j=0; i<e.pos.size(); i++,j+=5) {
		ElementPos ep = (ElementPos)e.pos.elementAt(i);
		addRow();
		((TextField)scr.getComponent(j)).setText(ep.time+"");
		((TextField)scr.getComponent(j+1)).setText(ep.x+"");
		((TextField)scr.getComponent(j+2)).setText(ep.y+"");
		((Checkbox)scr.getComponent(j+3)).setState(ep.visible);
		}
	addRow();	// one blank row
	}

	private void addRow()
	{
	Checkbox v;
	scr.add(new TextField(""));
	scr.add(new TextField(""));
	scr.add(new TextField(""));
	scr.add(v = new Checkbox("")); v.setState(true);
	scr.add(new CbButton("Go", this));
	}

	public void moved(CbScrollbar s, int p)
	{
	scr.move(0, -p);
	}

	public void moving(CbScrollbar s, int p)
	{
	}

	public boolean action(Event evt, Object obj)
	{
	// return hit somewhere... maybe delete row?
	Component c[] = scr.getComponents();
	for(int i=0; i<c.length-5; i+=5) {
		if (blank(c[i]) && blank(c[i+1]) && blank(c[i+2])) {
			scr.remove(c[i]); scr.remove(c[i+1]);
			scr.remove(c[i+2]); scr.remove(c[i+3]);
			scr.remove(c[i+4]);
			resize(size().width, size().height);
			scr.layout();
			return true;
			}
		}

	// .. or maybe add row?
	if (!blank(c[c.length-5]) && !blank(c[c.length-4]) &&
	    !blank(c[c.length-3])) {
		addRow();
		resize(size().width, size().height); scr.layout();
		}
	return true;
	}

	private boolean blank(Component c)
	{
	return ((TextField)c).getText().trim().equals("");
	}

	/**Returns a new vector of element positions, or null in case of
	 * error.
	 */
	Vector getPositions()
	{
	Vector rv = new Vector();
	Component c[] = scr.getComponents();
	for(int i=0; i<c.length; i+=5) {
		boolean b1 = blank(c[i]), b2 = blank(c[i+1]),
		        b3 = blank(c[i+2]);
		if (b1 && b2 && b3) continue;
		if (b1) {
			new ErrorWindow("Missing time from row "+i);
			return null;
			}
		if (b2) {
			new ErrorWindow("Missing X coordinate from row "+i);
			return null;
			}
		if (b3) {
			new ErrorWindow("Missing Y coordinate from row "+i);
			return null;
			}
		TextField t = (TextField)c[i],
			    x = (TextField)c[i+1],
			    y = (TextField)c[i+2];
		Checkbox v = (Checkbox)c[i+3];
		try {
			rv.addElement(new ElementPos(
			  Integer.parseInt(t.getText()),
			  Integer.parseInt(x.getText()),
			  Integer.parseInt(y.getText()),
			  v.getState()));
			}
		catch(NumberFormatException ex) {
			new ErrorWindow("Time, X and Y coordinates must be "+
			                "numeric in row "+i);
			return null;
			} 
		}
	if (rv.size() == 0) {
		new ErrorWindow("You must enter at least one position");
		return null;
		}
	return rv;
	}

	public void click(CbButton b)
	{
	Component c[] = scr.getComponents();
	for(int i=0; i<c.length; i+=5) {
		if (c[i+4] == b) {
			try {
				parent.edit_area.parent.gotoTime(
				 Integer.parseInt(((TextField)c[i]).getText()));
				}
			catch(Exception e) { }
			break;
			}
		}
	}

	public void reshape(int nx, int ny, int nw, int nh)
	{
	super.reshape(nx, ny, nw, nh);
	int sh = scr.minimumSize().height;
	Dimension midcs = midc.size();
	if (midcs.height <= 0) return;
	scr.resize(midcs.width, sh);
	int v = sb.getValue(), mv = sh-midcs.height;
	sb.setValues(mv<0 ? 0 : v > mv ? mv : v, midcs.height, sh);
	}

	public Dimension minimumSize()
	{
	return new Dimension(250, 100);
	}

	public Dimension preferredSize()
	{
	return minimumSize();
	}
}
