import java.awt.image.*;
import java.util.*;
import java.io.*;
import java.awt.*;

/**A collection of name/value pairs. Each name is a string, and each value
 * a string, integer, float, byte array, Image or HData object.
 * Can be written to or read from a file
 */
class HData
{
	private Hashtable d = new Hashtable();

	HData() { }

	HData(FixedDataInputStream is) throws IOException
	{
	read(is);
	}

	void addString(String n, String v) { d.put(n, v); }
	void addInteger(String n, int v) { d.put(n, new Integer(v)); }
	void addFloat(String n, float v) { d.put(n, new Float(v)); }
	void addVector(String n, Vector v) { d.put(n, v); }
	void addImage(String n, Image v) { d.put(n, v); }
	void addHData(String n, HData v) { d.put(n, v); }
	void addPoint(String n, Point v) { d.put(n, v); }
	void addColor(String n, Color v) { d.put(n, v); }
	void addBoolean(String n, boolean v) { d.put(n, new Boolean(v)); }
	void addFont(String n, Font f) { d.put(n, f); }

	String getString(String n) { return (String)d.get(n); }
	int getInteger(String n) { return ((Integer)d.get(n)).intValue(); }
	float getFloat(String n) { return ((Float)d.get(n)).floatValue(); }
	Vector getVector(String n) { return (Vector)d.get(n); }
	Image getImage(String n) { return (Image)d.get(n); }
	HData getHData(String n) { return (HData)d.get(n); }
	Point getPoint(String n) { return (Point)d.get(n); }
	Color getColor(String n) { return (Color)d.get(n); }
	boolean getBoolean(String n) { return ((Boolean)d.get(n)).booleanValue(); }
	Font getFont(String n) { return (Font)d.get(n); }

	boolean contains(String n) { return d.containsKey(n); }

	void write(DataOutputStream out) throws IOException
	{
	out.writeInt(d.size());
	for(Enumeration e = d.keys(); e.hasMoreElements(); ) {
		String k = (String)e.nextElement();
		Object v = d.get(k);
		out.writeUTF(k);
		writeValue(v, out);
		}
	}

	private void writeValue(Object v, DataOutputStream out)
		throws IOException
	{
	if (v instanceof String) {
		// a unicode string
		out.writeByte(0);
		out.writeUTF((String)v);
		}
	else if (v instanceof Integer) {
		// a single integer
		out.writeByte(1);
		out.writeInt(((Integer)v).intValue());
		}
	else if (v instanceof Float) {
		// a single floating point
		out.writeByte(2);
		out.writeFloat(((Float)v).floatValue());
		}
	else if (v instanceof Vector) {
		// a vector of writable objects
		out.writeByte(3);
		Vector vv = (Vector)v;
		out.writeInt(vv.size());
		for(int j=0; j<vv.size(); j++)
			writeValue(vv.elementAt(j), out);
		}
	else if (v instanceof Image) {
		Image i = (Image)v;
		int w = Util.getWidth(i), h = Util.getHeight(i);
		int buf[] = new int[w*h];
		PixelGrabber g =
		  new PixelGrabber(i, 0, 0, w, h, buf, 0, w);
		try { g.grabPixels(); } catch(Exception e) { }

		out.writeByte(4);
		out.writeInt(w); out.writeInt(h);
		byte bbuf[] = new byte[buf.length*4];
		for(int j=0,k=0; j<buf.length; j++,k+=4) {
			bbuf[k] = (byte)((buf[j]&0xff000000)>>24);
			bbuf[k+1] = (byte)((buf[j]&0xff0000)>>16);
			bbuf[k+2] = (byte)((buf[j]&0xff00)>>8);
			bbuf[k+3] = (byte)((buf[j]&0xff));
			}
		out.write(bbuf, 0, bbuf.length);
		}
	else if (v instanceof HData) {
		out.writeByte(5);
		((HData)v).write(out);
		}
	else if (v instanceof Point) {
		out.writeByte(6);
		Point p = (Point)v;
		out.writeInt(p.x);
		out.writeInt(p.y);
		}
	else if (v instanceof Color) {
		out.writeByte(7);
		Color c = (Color)v;
		out.writeInt(c.getRed());
		out.writeInt(c.getGreen());
		out.writeInt(c.getBlue());
		}
	else if (v instanceof Boolean) {
		out.writeByte(8);
		out.writeBoolean(((Boolean)v).booleanValue());
		}
	else if (v instanceof Font) {
		out.writeByte(9);
		Font f = (Font)v;
		out.writeUTF(f.getName());
		out.writeInt(f.getStyle());
		out.writeInt(f.getSize());
		}
	}

	void read(FixedDataInputStream in) throws IOException
	{
	int n = in.readInt();
	for(int i=0; i<n; i++) {
		String name = in.readUTF2();
		d.put(name, readValue(in));
		}
	}

	private Object readValue(FixedDataInputStream in) throws IOException
	{
		int type = in.readByte();
		switch(type) {
		case 0:	return in.readUTF2();
		case 1:	return new Integer(in.readInt());
		case 2:	return new Float(in.readFloat());
		case 3:	int l = in.readInt();
			Vector v = new Vector();
			for(int j=0; j<l; j++)
				v.addElement(readValue(in));
			return v;
		case 4:	int w = in.readInt(), h =in.readInt();
			byte bbuf[] = new byte[w*h*4];
			in.readFully(bbuf);
			int buf[] = new int[w*h];
			for(int j=0,k=0; j<bbuf.length; j+=4,k++)
				buf[k] = (((int)bbuf[j]&0xff)<<24) +
				         (((int)bbuf[j+1]&0xff)<<16) +
				         (((int)bbuf[j+2]&0xff)<<8) +
				         (((int)bbuf[j+3]&0xff));
			return Util.createImage(
			  new MemoryImageSource(w, h, buf, 0, w));
		case 5: return new HData(in);
		case 6: return new Point(in.readInt(), in.readInt());
		case 7: return new Color(in.readInt(), in.readInt(),
		                         in.readInt());
		case 8: return new Boolean(in.readBoolean());
		case 9: return new Font(in.readUTF2(), in.readInt(),
		                        in.readInt());
		}
	return null;	// can't happen
	}
}

