import java.awt.*;
import java.applet.Applet;
import java.awt.image.*;
import java.util.*;
import java.io.*;

/**This class is the representation of a presentation used by the editor.
 * It consists of 1 or more TimeLines, each of which contains a number of
 * moving and interactive objects.
 */
class Presentation implements Runnable
{
	Vector tl = new Vector();
	TimeLine ctl;
	int w, h;
	int frame = 20;

	Component playc;
	Applet playapl;
	Graphics playg, dbufg;
	Thread th;
	Image dbuf;
	Button dummy;
	boolean beenpaused;
	Vector entexit = new Vector();
	Element inside;
	Action timed[];
	int timedc;
	Object glock = new Object();

	/**Create a presentation with one timeline
	 * @param len	Length of the timeline
	 */
	Presentation(int len, int wi, int he)
	{
	tl.addElement(ctl = new TimeLine(this, len, "Default"));
	w = wi; h = he;
	}

	/**Create this presentation by reading data from the given stream
	 */
	Presentation(InputStream is) throws IOException
	{
	this(new HData(new FixedDataInputStream(is)));
	}

	/**Create this presentation from the given data source
	 */
	Presentation(HData hd)
	{
	w = hd.getInteger("w"); h = hd.getInteger("h");
	Vector tv = hd.getVector("tl");
	for(int i=0; i<tv.size(); i++) {
		TimeLine nt = new TimeLine((HData)tv.elementAt(i), this);
		tl.addElement(nt);
		}
	ctl = (TimeLine)tl.elementAt(hd.getInteger("ctl"));
	}

	/**Save this entire presentation to the given timeline
	 */
	void save(OutputStream o) throws IOException
	{
	save().write(new DataOutputStream(o));
	}

	/**Create a data object from this presentation
	 */
	HData save()
	{
	HData hd = new HData();
	hd.addInteger("w", w); hd.addInteger("h", h);
	Vector tv = new Vector();
	int ctli = 0;
	for(int i=0; i<tl.size(); i++) {
		tv.addElement(((TimeLine)tl.elementAt(i)).save());
		if (ctl == tl.elementAt(i))
			ctli = i;
		}
	hd.addVector("tl", tv);
	hd.addInteger("ctl", ctli);
	return hd;
	}

	/**Find the element with the given name (if any)
	 */
	Element findElement(String n)
	{
	for(int i=0; i<tl.size(); i++) {
		TimeLine t = (TimeLine)tl.elementAt(i);
		for(int j=0; j<t.elmc; j++)
			if (t.elm[j].name.equals(n))
				return t.elm[j];
		}
	return null;
	}

	/**Find the topmost element at some position
	 */
	Element findElement(int x, int y)
	{
	Element ev[] = ctl.elm;
	for(int i=ctl.elmc-1; i>=0; i--) {
		Element e = ev[i];
		if (e.getStyle() == 0 && e.getBounds().inside(x, y))
			return e;
		}
	return null;
	}

	/**Give a newly created element a unique name
	 */
	void nameElement(Element e)
	{
	String c = e.getClass().getName(), nm;
	int n = 1;
	do { nm = c+n++; }
	while(findElement(nm) != null);
	e.name = nm;
	}

	/**Returns an array of all elements from all presentations
	 */
	Element[] listElements()
	{
	Vector v = new Vector();
	for(int i=0; i<tl.size(); i++) {
		TimeLine t = (TimeLine)tl.elementAt(i);
		for(int j=0; j<t.elmc; j++)
			v.addElement(t.elm[j]);
		}
	Element rv[] = new Element[v.size()];
	v.copyInto(rv);
	return rv;
	}

	/**This method should be called before the first play to set the
	 * double-buffer and graphics for playing this presentation
	 */
	void setup(Image d, Component c, Button b)
	{
	playc = c;
	playg = playc.getGraphics();
	dbuf = d;
	dbufg = dbuf.getGraphics();
	dummy = b;
	Component w = c;
	while(w != null) {
		if (w instanceof Applet) {
			playapl = (Applet)w;
			break;
			}
		w = w.getParent();
		}
	}

	/**Play this presentation from the start of the first timeline
	 */
	void play()
	{
	play((TimeLine)tl.elementAt(0), 0);
	}

	void pause()
	{
	if (th != null) {
		synchronized(glock) { th.suspend(); }
		}
	beenpaused = true;
	}

	void unpause()
	{
	if (th != null) th.resume();
	}

	void halt()
	{
	if (th != null) th.stop();
	}

	/**Start playing this timeline. This continues until
	 *  - The timeline ends, after which the final frame will be
	 *    permanenly displayed.
	 *  - Some user action or timeline event causes a switch to
	 *    another timeline
	 */
	void play(TimeLine t, int p)
	{
	// handle exiting of inside element
	if (inside != null)
		callActions(entexit, 3, inside);
	inside = null;

	// build list of enter/exit actions
	entexit.removeAllElements();
	for(int i=0; i<t.elmc; i++) {
		Vector av = t.elm[i].acts;
		for(int j=0; j<av.size(); j++) {
			Action a = (Action)av.elementAt(j);
			if (a.event == 2 || a.event == 3)
				entexit.addElement(a);
			}
		}
	inside = null;

	// build list of timed actions
	timed = new Action[t.acts.size()];
	timedc = 0;
	for(int i=0; i<t.acts.size(); i++) {
		Action a = (Action)t.acts.elementAt(i);
		if (a.event == 0)
			timed[timedc++] = a;
		}

	ctl = t;
	ctl.setTime(p);
	if (th != null) th.stop();
	th = new Thread(this);
	th.start();
	}

	/**Play this timeline from start to finish
	 */
	public void run()
	{
	long now = System.currentTimeMillis(), prev = now;
	while(ctl.time <= ctl.len) {
		try { Thread.sleep(frame); }
		catch(Exception e) { }
		int nt = ctl.time+(int)((now-prev) * ctl.speed);
		if (nt > ctl.len) nt = ctl.len;

		// Check for timed events
		for(int i=0; i<timedc; i++) {
			int t = timed[i].time;
			if (t == 0) t = 1;
			else if (t >= ctl.len) t = ctl.len-1;
			if (t > ctl.time && t <= nt)
				new TimedActionRunner(timed[i], this);
			}

		ctl.setTime(nt); paint();
		if (beenpaused) {
			prev = now = System.currentTimeMillis();
			beenpaused = false;
			}
		else {
			prev = now;
			now = System.currentTimeMillis();
			}
		dummy.setLabel(dummy.getLabel());
		}
	th = null;
	}

	/**Render into the current graphics */
	void paint()
	{
	synchronized(glock) {
		ctl.paintBackground(dbufg, false);
		ctl.paintElements(dbufg, false);
		playg.drawImage(dbuf, 0, 0, null);
		}
	}

	/**Called by actions if the presentation is changed and a repaint
	 * is needed */
	void changed()
	{
	ctl.setTime(ctl.time);
	paint();
	}

	void mouseClick(int x, int y)
	{
	Element ce = findElement(x, y);
	if (ce != null) callActions(ce.acts, 1, ce);
	else callActions(ctl.acts, 1, ctl);
	}

	void mouseMove(int x, int y)
	{
	if (entexit.size() == 0) {
		// no need to handle enter/exit events
		return;
		}
	// check all the MouseEnter and MouseExit actions
	Element ni = findElement(x, y);
	if (ni != inside) {
		// inside-ness has changed
		if (inside != null) callActions(entexit, 3, inside);
		if (ni != null) callActions(entexit, 2, ni);
		inside = ni;
		}
	}

	private void callActions(Vector v, int e, Object o)
	{
	for(int i=0; i<v.size(); i++) {
		Action a = (Action)v.elementAt(i);
		if (a.event == e && a.owner == o)
			a.activate(this);
		}
	}

}


class TimedActionRunner extends Thread
{
	Action act;
	Presentation pres;

	TimedActionRunner(Action a, Presentation p)
	{
	act = a;
	pres = p;
	start();
	}

	public void run()
	{
	synchronized(pres.glock) {
		act.activate(pres);
		}
	}
}

