import java.awt.*;
import java.net.*;
import java.util.*;

/**A Timeline is one screen or animation in a presentation. A timeline
 * contains a number of Elements, each of which can be animated.
 */
class TimeLine
{
	Presentation parent;
	String name;
	Element elm[] = new Element[10];
	int elmc;
	int len, time;
	float speed = 1;
	Color fillcol = Color.white;
	Image fillimg;
	String fillimgsrc;
	int fillimg_w, fillimg_h;
	boolean fillimg_tile;
	Vector acts = new Vector();

	/**Create a new timeline with the given parent, length and name
	 */
	TimeLine(Presentation p, int l, String n)
	{
	parent = p;
	len = l;
	name = n;
	}

	/**Create a timeline read from a previous save
	 */
	TimeLine(HData hd, Presentation p)
	{
	parent = p;
	name = hd.getString("name");
	len = hd.getInteger("len");
	speed = hd.getFloat("speed");
	Vector ev = hd.getVector("elm");
	fillcol = hd.getColor("fillcol");
	fillimg_tile = hd.getBoolean("fillimg_tile");
	if (hd.contains("fillimgsrc")) {
		fillimgsrc = hd.getString("fillimgsrc");
		try {
			URL u = new URL(fillimgsrc);
			fillimg = Util.tk.getImage(u);
			}
		catch(Exception e) { }
		if (fillimg == null)
			new ErrorWindow("Failed to load image from "+
			                fillimgsrc+" in timeline "+name);
		}
	else if (hd.contains("fillimg"))
		fillimg = hd.getImage("fillimg");
	if (fillimg != null) {
		fillimg_w = Util.getWidth(fillimg);
		fillimg_h = Util.getHeight(fillimg);
		if (fillimg_tile) Util.waitForImage(fillimg);
		else Util.waitForImage(fillimg, parent.w, parent.h);
		}
	elmc = ev.size(); elm = new Element[elmc];
	for(int i=0; i<ev.size(); i++) {
		HData ehd = (HData)ev.elementAt(i);
		Element ne = (Element)Util.createObject(ehd.getString("class"));
		ne.load(ehd);
		ne.parent = this;
		elm[ne.idx = i] = ne;
		}
	setTime(hd.getInteger("time"));
	Vector av = hd.getVector("acts");
	for(int i=0; i<av.size(); i++) {
		Action ac = Action.loadAction((HData)av.elementAt(i));
		ac.owner = this;
		acts.addElement(ac);
		}
	}

	void setLength(int l) { len = l; }

	void setName(String n) { name = n; }

	void setTime(int t)
	{
	time = t;
	for(int i=0; i<elmc; i++)
		elm[i].setTime(t);
	}

	int numElements() { return elmc; }

	void addElement(Element e)
	{
	if (elmc == elm.length) {
		Element nelm[] = new Element[elmc*2 + 1];
		System.arraycopy(elm, 0, nelm, 0, elmc);
		elm = nelm;
		}
	elm[e.idx = elmc++] = e;
	e.parent = this;
	}

	void removeElement(int p)
	{
	for(int i=p+1; i<elmc; i++) {
		elm[i-1] = elm[i];
		elm[i-1].idx--;
		}
	elmc--;
	}

	/**Deletes an element
	 * @param e		The element to delete
	 */
	void removeElement(Element e) { removeElement(e.idx); }

	/**Returns the element at some position
	 * @param p		The element to return
	 */
	Element getElement(int p)
	{
	return elm[p];
	}

	/**Move an element from one position to another, shifting other
	 * elements up or down as needed
	 * @param from	Position of element to move
	 * @param to	Position to move to
	 */
	void moveElement(int from, int to)
	{
	Element e = elm[from];
	if (to < from) {
		for(int i=from; i>to; i--) {
			elm[i] = elm[i-1];
			elm[i].idx++;
			}
		}
	else if (from < to) {
		for(int i=from; i<to; i++) {
			elm[i] = elm[i+1];
			elm[i].idx--;
			}
		}
	elm[e.idx = to] = e;
	}

	/**Move an element from one position to another, shifting other
	 * elements up or down as needed
	 * @param e		Element to move
	 * @param to	Position to move to
	 */
	void moveElement(Element e, int to) { moveElement(e.idx, to); }

	/**Render all objects in this timeline into the given graphics
	 * at the current time.
	 * @param g		Graphics to draw into
	 * @param e		Are we in edit mode?
	 */
	void paintElements(Graphics g, boolean e)
	{
	for(int i=0; i<elmc; i++)
		if (elm[i].visible)
			elm[i].paint(g, e);
	}

	/**Render the background of this timeline into the given graphics
	 * @param g	Graphics to draw into
	 * @param e		Are we in edit mode?
	 */
	void paintBackground(Graphics g, boolean e)
	{
	if (fillcol != null) {
		g.setColor(fillcol);
		g.fillRect(0, 0, parent.w, parent.h);
		}
	if (fillimg != null) {
		if (fillimg_tile) {
			for(int i=0; i<parent.w; i+=fillimg_w)
				for(int j=0; j<parent.h; j+=fillimg_h)
					g.drawImage(fillimg, i, j, Util.fr);
			}
		else
			g.drawImage(fillimg, 0, 0, parent.w, parent.h, Util.fr);
		}
	}

	/**Save this timeline and all elements into a data object
	 */
	HData save()
	{
	HData hd = new HData();
	hd.addString("name", name);
	hd.addInteger("len", len);
	hd.addInteger("time", time);
	hd.addFloat("speed", speed);
	hd.addColor("fillcol", fillcol);
	if (fillimg != null) {
		if (fillimgsrc != null) hd.addString("fillimgsrc", fillimgsrc);
		else hd.addImage("fillimg", fillimg);
		}
	hd.addBoolean("fillimg_tile", fillimg_tile);
	Vector ev = new Vector();
	for(int i=0; i<elmc; i++)
		ev.addElement(elm[i].save());
	hd.addVector("elm", ev);
	Vector av = new Vector();
	for(int i=0; i<acts.size(); i++)
		av.addElement(((Action)acts.elementAt(i)).save());
	hd.addVector("acts", av);
	return hd;
	}

	boolean mouseDown(Event e, int x, int y)
	{
	for(int i=elmc-1; i>=0; i--) {
		boolean rv = elm[i].mouseDown(e, x, y);
		if (rv) return true;
		}
	return false;
	}

	boolean mouseUp(Event e, int x, int y)
	{
	return false;
	}

	boolean mouseDrag(Event e, int x, int y)
	{
	return false;
	}
 }
